/* ***********************************************************************/
/*                                                                       */
/*                     WiNRADiO G69DDC EXCELSIOR ULTRA                   */
/*                         G69DDC device driver                          */
/*                      (c) Copyright 2022 RADIXON.                      */
/*                                                                       */
/* ***********************************************************************/

#include "g69ddc.h"


static struct class g69ddc_class=
{
    .name=G69DDC_MODULE_NAME
};

static dev_t base_device_number;
static struct mutex reg_mutex;
static struct g69ddc *registered_devices[G69DDC_MAX_DEVICE_COUNT];

static inline void reg_lock(void)
{
    mutex_lock(&reg_mutex);
}

static inline void reg_unlock(void)
{
    mutex_unlock(&reg_mutex);
}

static inline void dev_lock(struct g69ddc *dev)
{
    mutex_lock(&dev->mutex);
}

static inline void dev_unlock(struct g69ddc *dev)
{
    mutex_unlock(&dev->mutex);
}

#define SET_STATE(__dev__,__operation__) (__dev__)->state.s.__operation__;wake_up(&(__dev__)->state_wait_queue)

static int g69ddc_fo_open(struct inode *node, struct file *file);
static int g69ddc_fo_release(struct inode *node, struct file *file);

static long g69ddc_fo_ioctl(struct file *file, unsigned int cmd, unsigned long arg);


#ifdef G69DDC_COMPAT
static long g69ddc_fo_ioctl_compat(struct file *file, unsigned int cmd, unsigned long arg);
#endif

static struct file_operations file_operations=
{
    .open=g69ddc_fo_open,
    .release=g69ddc_fo_release,
    .unlocked_ioctl=g69ddc_fo_ioctl,

#ifdef G69DDC_COMPAT
    .compat_ioctl=g69ddc_fo_ioctl_compat,
#endif
};


static inline void g69ddc_queue_wd_work(struct g69ddc *dev)
{
    queue_delayed_work(dev->workqueue,&dev->wd_delayed_work,HZ*2);
}

static inline void g69ddc_cancel_wd_work(struct g69ddc *dev)
{
    cancel_delayed_work_sync(&dev->wd_delayed_work);
}

static void g69ddc_wd(struct work_struct *work)
{
 struct g69ddc *dev;
 bool critical_temperature;
 int status;

    dev=container_of(work,struct g69ddc,wd_delayed_work.work);

    dev_lock(dev);

    if(dev->state.s.power && dev->state.s.connected && dev->open)
    {
        status=dev->fnc_check_temperature_locked(dev,&critical_temperature);

        for(;;)
        {
            if(likely(status==0))
            {
                if(unlikely(critical_temperature))
                {
                    dev->wd_counter++;

                    if(dev->wd_counter>=15)
                    {
                        dev_print_error(dev->p_dev,"critical temperature detected!");

                        dev->state.s.critical_temperature=1;
                        dev->fnc_set_power_locked(dev,0);

                        dev_print_warn(dev->p_dev,"device has been turned off!");
                        break;
                    }
                }
                else
                {
                    dev->wd_counter=0;
                }
            }
            else
            {
                dev->fnc_set_power_locked(dev,0);
                break;
            }

            g69ddc_queue_wd_work(dev);
            break;
        }
    }

    dev_unlock(dev);
}

static int g69ddc_initialize(struct g69ddc *dev)
{
    dev->workqueue=create_singlethread_workqueue(G69DDC_MODULE_NAME);

    if(unlikely(dev->workqueue==NULL))
    {
        dev_print_error(dev->p_dev,"unable to create work queue!");
        return -ENOMEM;
    }

    INIT_DELAYED_WORK(&dev->wd_delayed_work,g69ddc_wd);

    mutex_init(&dev->mutex);
    init_waitqueue_head(&dev->state_wait_queue);

    dev->state.s.connected=1;

    return 0;
}


static void g69ddc_uninitialize(struct g69ddc *dev)
{
    if(dev->fnc_uninitialize)
    {
        dev->fnc_uninitialize(dev);
    }

    destroy_workqueue(dev->workqueue);

    mutex_destroy(&dev->mutex);
}

static void g69ddc_delete(struct g69ddc *dev)
{
    flush_workqueue(dev->workqueue);

    if(dev->fnc_delete)
    {
        dev->fnc_delete(dev);
    }

    g69ddc_uninitialize(dev);

    kfree(dev);
}

static void release_callback(struct device *dev)
{
 struct g69ddc *g69ddc_dev;

    g69ddc_dev=container_of(dev,struct g69ddc,dev);
    g69ddc_delete(g69ddc_dev);
}

static int g69ddc_register(struct g69ddc *dev)
{
 int status;
 int index;
 dev_t dev_no;

    reg_lock();

    for(index=0;index<G69DDC_MAX_DEVICE_COUNT;index++)
    {
        if(registered_devices[index]==NULL)
        {
            break;
        }
    }

    if(index>=G69DDC_MAX_DEVICE_COUNT)
    {
        reg_unlock();

        dev_print_error(dev->p_dev,"no free device slot available!\n");
        return -ENFILE;
    }

    dev->index=index;

    dev_no=MKDEV(MAJOR(base_device_number),index);

    cdev_init(&dev->cdev,&file_operations);

    dev->cdev.owner=THIS_MODULE;

    status=cdev_add(&dev->cdev,dev_no,1);

    if(status)
    {
        reg_unlock();
        dev_print_error(dev->p_dev,"char device registration failed!");
        return status;
    }

    memset(&dev->dev,0,sizeof(dev->dev));

    dev->dev.devt=dev_no;
    dev->dev.class=&g69ddc_class;
    dev->dev.release=release_callback;

    dev_set_name(&dev->dev,G69DDC_MODULE_NAME"%d",index);


    status=device_register(&dev->dev);

    if(status)
    {
        cdev_del(&dev->cdev);
        reg_unlock();
        dev_print_error(dev->p_dev,"failed to register device!");
        return status;
    }

    registered_devices[index]=dev;

    reg_unlock();

    return 0;
}


static void g69ddc_unregister(struct g69ddc *dev)
{
    reg_lock();

    if(registered_devices[dev->index]==dev)
    {
        registered_devices[dev->index]=NULL;

        cdev_del(&dev->cdev);
        device_unregister(&dev->dev);
    }

    reg_unlock();
}

static int g69ddc_ioctl_get_info(struct g69ddc *dev,void __user *user_ptr)
{
 int i;
 struct g69ddc_ioctl_info info;

    info.driver_version=INTERNAL_DRIVER_VERSION;
    info.bus_type=dev->bus_type;

    for(i=0;i<4;i++)
    {
        info.fw[i]=dev->fw[i];
        info.vid_pid[i]=dev->vid_pid[i];
    }

    if(unlikely(copy_to_user(user_ptr,&info,sizeof(info))))
    {
        return -EFAULT;
    }

    return 0;
}

static int g69ddc_ioctl_get_state(struct g69ddc *dev,struct file *file,void __user *user_ptr)
{
 union g69ddc_state state;
 int status;
 DEFINE_WAIT(wait);

    if(copy_from_user(&state,user_ptr,sizeof(state)))
    {
        return -EFAULT;
    }

    dev_lock(dev);

    for(;;)
    {
        if(state.value!=dev->state.value)
        {
            if(unlikely(copy_to_user(user_ptr,&dev->state,sizeof(state))))
            {
                status=-EFAULT;
            }
            else
            {
                status=0;
            }
            break;
        }

        if(!dev->open)
        {
            status=-EPERM;
            break;
        }

        if(file->f_flags & O_NONBLOCK)
        {
            status=-EAGAIN;
            break;
        }

        prepare_to_wait(&dev->state_wait_queue,&wait,TASK_INTERRUPTIBLE);

        dev_unlock(dev);

        schedule();

        dev_lock(dev);

        finish_wait(&dev->state_wait_queue,&wait);

        if(unlikely(signal_pending(current)))
        {
            status=-ERESTARTSYS;
            break;
        }
    }

    dev_unlock(dev);

    return status;
}

static int g69ddc_ioctl_set_power(struct g69ddc *dev,uint8_t power)
{
 int status;

    dev_lock(dev);

    if(power && dev->state.s.power==0)
    {
        dev->state.s.critical_temperature=0;
        dev->state.s.suspended=0;
        dev->wd_counter=0;
    }

    status=dev->fnc_set_power_locked(dev,power);

    if(power)
    {
        g69ddc_queue_wd_work(dev);
    }

    dev_unlock(dev);

    if(!power)
    {
        g69ddc_cancel_wd_work(dev);
    }

    return status;
}

static int g69ddc_ioctl(struct g69ddc *dev,struct file *file, unsigned int cmd,unsigned long arg)
{
    switch(cmd)
    {
        case G69DDC_IOCTL_GET_STATE:
            return g69ddc_ioctl_get_state(dev,file,(void __user*)arg);

        case G69DDC_IOCTL_SET_POWER:
            return g69ddc_ioctl_set_power(dev,(uint8_t)arg);

        case G69DDC_IOCTL_GET_INFO:
            return g69ddc_ioctl_get_info(dev,(void __user*)arg);

        default:
            return -ENOIOCTLCMD;
    }
}

#ifdef G69DDC_COMPAT

static int g69ddc_ioctl_compat(struct g69ddc *dev,struct file *file, unsigned int cmd,unsigned long arg)
{
    switch(cmd)
    {
        case G69DDC_IOCTL_GET_STATE:
            return g69ddc_ioctl_get_state(dev,file,compat_ptr(arg));

        case G69DDC_IOCTL_SET_POWER:
            return g69ddc_ioctl_set_power(dev,(uint8_t)arg);

        case G69DDC_IOCTL_GET_INFO:
            return g69ddc_ioctl_get_info(dev,compat_ptr(arg));

        default:
            return -ENOIOCTLCMD;
    }
}

#endif //G69DDC_COMPAT

#ifdef G69DDC_USB

static inline void dev_usb_lock(struct g69ddc_usb *dev)
{
    dev_lock(&dev->base_dev);
}

static inline void dev_usb_unlock(struct g69ddc_usb *dev)
{
    dev_unlock(&dev->base_dev);
}

static inline void usb_lock(struct g69ddc_usb *dev)
{
    down_read(&dev->rw_sem);
}

static inline void usb_unlock(struct g69ddc_usb *dev)
{
    up_read(&dev->rw_sem);
}

static inline void usb_lock_exclusive(struct g69ddc_usb *dev)
{
    down_write(&dev->rw_sem);
}

static inline void usb_unlock_exclusive(struct g69ddc_usb *dev)
{
    up_write(&dev->rw_sem);
}

static void put_pages(struct page **pages,int count)
{
 int i;

    for(i=0;i<count;i++)
    {
        put_page(pages[i]);
    }
}


static int g69ddc_usb_ep_add_buffer(struct g69ddc_usb *dev,uint32_t ep_index,uintptr_t user_buffer,uint32_t size)
{
 struct g69ddc_usb_ep_buffer *buffer;
 int page_count,pinned,i;
 unsigned long flags;

    if(unlikely(user_buffer==0 || size==0 || (user_buffer & (~PAGE_MASK))!=0 || ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    page_count=DIV_ROUND_UP(size,PAGE_SIZE);

    usb_lock(dev);

    if(unlikely(dev->usb_interface==NULL))
    {
        usb_unlock(dev);
        return -ENODEV;
    }

    buffer=kmalloc(sizeof(struct g69ddc_usb_ep_buffer),GFP_KERNEL);

    if(unlikely(buffer==NULL))
    {
        usb_unlock(dev);
        dev_print_error(&dev->usb_dev->dev,"out of memory!");
        return -ENOMEM;
    }

    buffer->pages=kmalloc(page_count*sizeof(struct page*),GFP_KERNEL);

    if(unlikely(buffer->pages==NULL))
    {
        usb_unlock(dev);
        kfree(buffer);
        dev_print_error(&dev->usb_dev->dev,"out of memory!");
        return -ENOMEM;
    }

    buffer->sg=kmalloc(page_count*sizeof(struct scatterlist),GFP_KERNEL);

    if(unlikely(buffer->sg==NULL))
    {
        usb_unlock(dev);
        kfree(buffer->pages);
        kfree(buffer);
        dev_print_error(&dev->usb_dev->dev,"out of memory!");
        return -ENOMEM;
    }

    buffer->urb=usb_alloc_urb(0,GFP_KERNEL);

    if(unlikely(buffer->urb==NULL))
    {
        usb_unlock(dev);
        kfree(buffer->sg);
        kfree(buffer->pages);
        kfree(buffer);
        dev_print_error(&dev->usb_dev->dev,"out of memory!");
        return -ENOMEM;
    }

    pinned=get_user_pages_fast(user_buffer,page_count,usb_pipein(dev->ep[ep_index].pipe)!=0,buffer->pages);

    if(unlikely(pinned<0))
    {
        usb_unlock(dev);
        usb_free_urb(buffer->urb);
        kfree(buffer->sg);
        kfree(buffer->pages);
        kfree(buffer);
        return -EFAULT;
    }

    if(unlikely(pinned<page_count))
    {
        usb_unlock(dev);
        put_pages(buffer->pages,pinned);
        usb_free_urb(buffer->urb);
        kfree(buffer->sg);
        kfree(buffer->pages);
        kfree(buffer);
        return -EFAULT;
    }

    buffer->vmem=vmap(buffer->pages,pinned,0,PAGE_KERNEL);

    if(unlikely(buffer->vmem==NULL))
    {
        usb_unlock(dev);
        put_pages(buffer->pages,pinned);
        usb_free_urb(buffer->urb);
        kfree(buffer->sg);
        kfree(buffer->pages);
        kfree(buffer);
        return -ENOMEM;
    }


    buffer->user_buffer=user_buffer;
    buffer->size=size;
    buffer->page_count=pinned;
    buffer->status=0;
    buffer->idle=true;
    buffer->bytes_transferred=0;
    buffer->ep=&dev->ep[ep_index];

    sg_init_table(buffer->sg,page_count);

    for(i=0;i<page_count;i++)
    {
        if(size>=PAGE_SIZE)
        {
            sg_set_page(&buffer->sg[i],buffer->pages[i],PAGE_SIZE,0);
            size-=PAGE_SIZE;
        }
        else
        {
            sg_set_page(&buffer->sg[i],buffer->pages[i],size,0);
            break;
        }
    }


    spin_lock_irqsave(&dev->ep[ep_index].spinlock,flags);

    list_add_tail(&buffer->list_entry,&dev->ep[ep_index].buffer_list);

    spin_unlock_irqrestore(&dev->ep[ep_index].spinlock,flags);

    usb_unlock(dev);

    return 0;
}

static inline int g69ddc_usb_ioctl_ep_add_buffer(struct g69ddc_usb *dev,void __user *user_ptr)
{
 struct g69ddc_ioctl_usb_add_ep_buffer tmp;

    if(unlikely(copy_from_user(&tmp,user_ptr,sizeof(tmp))))
    {
        return -EFAULT;
    }

    return g69ddc_usb_ep_add_buffer(dev,tmp.ep_index,(uintptr_t)tmp.buffer,tmp.size);
}

static void g69ddc_usb_ep_cancel_buffers(struct g69ddc_usb *dev,struct g69ddc_usb_ep *ep)
{
    usb_lock(dev);

    if(unlikely(dev->usb_interface!=NULL))
    {
        usb_kill_anchored_urbs(&ep->anchor);
    }

    usb_unlock(dev);
}

static int g69ddc_usb_ioctl_ep_cancel_buffers(struct g69ddc_usb *dev,uint32_t ep_index)
{
    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    g69ddc_usb_ep_cancel_buffers(dev,&dev->ep[ep_index]);

    return 0;
}

static void g69ddc_usb_ep_release_buffer(struct g69ddc_usb_ep_buffer *buffer)
{
    vunmap(buffer->vmem);
    put_pages(buffer->pages,buffer->page_count);
    kfree(buffer->sg);
    kfree(buffer->pages);
    usb_free_urb(buffer->urb);
    kfree(buffer);
}

static void g69ddc_usb_ep_release_buffers(struct g69ddc_usb_ep *ep)
{
 unsigned long flags;
 struct g69ddc_usb_ep_buffer *buffer;
 struct list_head buffer_list;

    INIT_LIST_HEAD(&buffer_list);

    spin_lock_irqsave(&ep->spinlock,flags);

    list_splice_init(&ep->buffer_list,&buffer_list);

    spin_unlock_irqrestore(&ep->spinlock,flags);

    list_for_each_entry(buffer,&buffer_list,list_entry)
    {
        g69ddc_usb_ep_release_buffer(buffer);
    }
}

static int g69ddc_usb_ioctl_ep_cancel_and_release_buffers(struct g69ddc_usb *dev,uint32_t ep_index)
{
 struct g69ddc_usb_ep *ep;

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    ep=dev->ep+ep_index;

    g69ddc_usb_ep_cancel_buffers(dev,ep);
    g69ddc_usb_ep_release_buffers(ep);

    return 0;
}

static void g69ddc_usb_urb_complete(struct urb *urb)
{
 struct g69ddc_usb_ep_buffer *buffer;
 struct g69ddc_usb_ep *ep;
 unsigned long flags;

    buffer=urb->context;
    ep=buffer->ep;

    spin_lock_irqsave(&ep->spinlock,flags);

    buffer->idle=true;
    buffer->status=urb->status;

    if(likely(buffer->status==0))
    {
        buffer->bytes_transferred=urb->actual_length;
    }

    if(list_first_entry(&ep->buffer_list,typeof(*buffer),list_entry)==buffer)
    {
        wake_up(&ep->wait_queue);
    }

    spin_unlock_irqrestore(&ep->spinlock,flags);
}

static int g69ddc_usb_ioctl_ep_bulk_requeue_buffer(struct g69ddc_usb *dev,void __user *user_ptr)
{
 struct g69ddc_ioctl_usb_ep_requeue_buffer requeue_buffer;
 struct g69ddc_usb_ep *ep;
 struct g69ddc_usb_ep_buffer *buffer;
 unsigned long flags;
 int status;

    if(unlikely(copy_from_user(&requeue_buffer,user_ptr,sizeof(requeue_buffer))))
    {
        return -EFAULT;
    }

    if(unlikely(requeue_buffer.ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    ep=&dev->ep[requeue_buffer.ep_index];

    if(unlikely(ep->type!=USB_ENDPOINT_XFER_BULK))
    {
        return -EINVAL;
    }

    if(unlikely(requeue_buffer.transfer_size<G69DDC_USB_EP_MIN_TRANSFER_SIZE))
    {
        requeue_buffer.transfer_size=G69DDC_USB_EP_MIN_TRANSFER_SIZE;
    }
    else if(unlikely(requeue_buffer.transfer_size>G69DDC_USB_EP_MAX_TRASNFER_SIZE))
    {
        requeue_buffer.transfer_size=G69DDC_USB_EP_MAX_TRASNFER_SIZE;
    }

    usb_lock(dev);

    if(unlikely(dev->usb_interface==NULL))
    {
        usb_unlock(dev);
        return -ENODEV;
    }

    spin_lock_irqsave(&ep->spinlock,flags);

    if(unlikely(list_empty(&ep->buffer_list)))
    {
        spin_unlock_irqrestore(&ep->spinlock,flags);
        usb_unlock(dev);
        return 0;
    }

    buffer=list_first_entry(&ep->buffer_list,typeof(*buffer),list_entry);

    if(unlikely(!buffer->idle))
    {
        spin_unlock_irqrestore(&ep->spinlock,flags);
        usb_unlock(dev);
        return 0;
    }

    list_move_tail(&buffer->list_entry,&ep->buffer_list);

    buffer->idle=false;
    buffer->bytes_transferred=0;
    buffer->status=0;

    requeue_buffer.transfer_size=min(buffer->size,requeue_buffer.transfer_size);

    usb_fill_bulk_urb(buffer->urb,dev->usb_dev,ep->pipe,NULL,requeue_buffer.transfer_size,
                              g69ddc_usb_urb_complete,buffer);

    buffer->urb->sg=buffer->sg;
    buffer->urb->num_sgs=buffer->page_count;

    usb_anchor_urb(buffer->urb,&ep->anchor);

    status=usb_submit_urb(buffer->urb,GFP_KERNEL);

    if(unlikely(status))
    {
        buffer->idle=true;
        buffer->status=status;
        usb_unanchor_urb(buffer->urb);
    }

    spin_unlock_irqrestore(&ep->spinlock,flags);

    usb_unlock(dev);

    return status;
}

static inline int g69ddc_usb_set_led_pattern(struct g69ddc_usb *dev,uint16_t pattern)
{
 int status;

    usb_lock(dev);

    if(likely(dev->usb_interface))
    {
        status=usb_control_msg(dev->usb_dev,dev->snd_ctrl_pipe,
                           0xC0,USB_TYPE_VENDOR|USB_DIR_OUT,pattern,0,NULL,0,G69DDC_USB_TIMEOUT);
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock(dev);

    return status;
}

static int g69ddc_usb_update_led_pattern(struct g69ddc_usb *dev)
{
 uint16_t pattern;

    switch(dev->led_mode)
    {
        case USB_LED_MODE_OFF:
            pattern=USB_LED_PATTERN_OFF;
            break;

        case USB_LED_MODE_ON:
            pattern=USB_LED_PATTERN_ON;
            break;

        default:
            if(dev->base_dev.open)
            {
                pattern=dev->base_dev.state.s.power?USB_LED_PATTERN_POWER_ON:USB_LED_PATTERN_POWER_OFF;
            }
            else
            {
                pattern=USB_LED_PATTERN_APP_NOT_RUNNING;
            }
            break;
    }

    return g69ddc_usb_set_led_pattern(dev,pattern);
}

static inline int g69ddc_usb_clear_halt(struct g69ddc_usb *dev,struct g69ddc_usb_ep *ep)
{
 int status;

    usb_lock(dev);

    if(likely(dev->usb_interface))
    {
        status=usb_clear_halt(dev->usb_dev,ep->pipe);
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock(dev);

    return status;
}

static int g69ddc_usb_ioctl_clear_halt(struct g69ddc_usb *dev,uint32_t ep_index)
{
 int status;

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    usb_lock(dev);

    if(likely(dev->usb_interface))
    {
        status=usb_clear_halt(dev->usb_dev,dev->ep[ep_index].pipe);
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock(dev);

    return status;
}


static int g69ddc_usb_ioctl_reset_ep(struct g69ddc_usb *dev,uint32_t ep_index)
{
 int status;

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    usb_lock(dev);

    if(likely(dev->usb_interface))
    {
        usb_reset_endpoint(dev->usb_dev,dev->ep[ep_index].address);

        status=0;
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock(dev);

    return status;
}

static int g69ddc_usb_ctrl_vendor_msg_snd(struct g69ddc_usb *dev,uint8_t request,uint16_t value,uint16_t index,void __user *user_buffer,uint16_t size)
{
 void *buffer;
 int ret;

    if(unlikely((user_buffer!=NULL) ^ (size!=0)))
    {
        return -EINVAL;
    }

    if(user_buffer!=NULL)
    {
        buffer=kmalloc(size,GFP_KERNEL);

        if(unlikely(buffer==NULL))
        {
            return -ENOMEM;
        }

        if(unlikely(copy_from_user(buffer,user_buffer,size)))
        {
            kfree(buffer);
            return -EFAULT;
        }
    }
    else
    {
        buffer=NULL;
    }

    usb_lock(dev);

    if(likely(dev->usb_interface))
    {
        ret=usb_control_msg(dev->usb_dev,dev->snd_ctrl_pipe,request,USB_TYPE_VENDOR|USB_DIR_OUT,value,index,buffer,size,G69DDC_USB_TIMEOUT);
    }
    else
    {
        ret=-ENODEV;
    }

    usb_unlock(dev);

    if(buffer!=NULL)
    {
        kfree(buffer);
    }

    return ret;
}

static int g69ddc_usb_ctrl_vendor_msg_rcv(struct g69ddc_usb *dev,uint8_t request,uint16_t value,uint16_t index,void __user *user_buffer,uint16_t size)
{
 void *buffer;
 int ret;

    if(unlikely((user_buffer!=NULL) ^ (size!=0)))
    {
        return -EINVAL;
    }

    if(user_buffer!=NULL)
    {
        buffer=kmalloc(size,GFP_KERNEL);

        if(unlikely(buffer==NULL))
        {
            return -ENOMEM;
        }
    }
    else
    {
        buffer=NULL;
    }

    usb_lock(dev);

    if(likely(dev->usb_interface))
    {
        ret=usb_control_msg(dev->usb_dev,dev->rcv_ctrl_pipe,request,USB_TYPE_VENDOR|USB_DIR_IN,value,index,buffer,size,G69DDC_USB_TIMEOUT);
    }
    else
    {
        ret=-ENODEV;
    }

    usb_unlock(dev);

    if(buffer!=NULL)
    {
        if(ret>0)
        {
            if(unlikely(copy_to_user(user_buffer,buffer,ret)))
            {
                ret=-EFAULT;
            }
        }

        kfree(buffer);
    }

    return ret;
}

static inline int g69ddc_usb_ioctl_ctrl_vendor_msg_snd(struct g69ddc_usb *dev,void __user *user_ptr)
{
 struct g69ddc_ioctl_usb_ctrl_vendor_msg msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g69ddc_usb_ctrl_vendor_msg_snd(dev,msg.request,msg.value,msg.index,(void __user*)msg.data,msg.size);
}

static inline int g69ddc_usb_ioctl_ctrl_vendor_msg_rcv(struct g69ddc_usb *dev,void __user *user_ptr)
{
 struct g69ddc_ioctl_usb_ctrl_vendor_msg msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g69ddc_usb_ctrl_vendor_msg_rcv(dev,msg.request,msg.value,msg.index,(void __user*)msg.data,msg.size);
}

static int g69ddc_usb_bulk_msg_snd(struct g69ddc_usb *dev,uint32_t ep_index,void __user *user_buffer,uint32_t size)
{
 void *buffer;
 int transferred;
 int status;
 struct g69ddc_usb_ep *ep;

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    ep=dev->ep+ep_index;

    if(unlikely(user_buffer==NULL || size==0 || (ep->address&USB_DIR_IN) || ep->type!=USB_ENDPOINT_XFER_BULK))
    {
        return -EINVAL;
    }

    buffer=kmalloc(size,GFP_KERNEL);

    if(unlikely(buffer==NULL))
    {
        return -ENOMEM;
    }

    if(unlikely(copy_from_user(buffer,user_buffer,size)))
    {
        kfree(buffer);
        return -EFAULT;
    }

    usb_lock(dev);

    if(likely(dev->usb_interface))
    {
        status=usb_bulk_msg(dev->usb_dev,ep->pipe,buffer,size,&transferred,G69DDC_USB_TIMEOUT);
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock(dev);

    kfree(buffer);

    if(unlikely(status<0))
    {
        return status;
    }

    return transferred;
}


static int g69ddc_usb_bulk_msg_rcv(struct g69ddc_usb *dev,uint32_t ep_index,void __user *user_buffer,uint32_t size)
{
 void *buffer;
 int transferred;
 int status;
 struct g69ddc_usb_ep *ep;

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    ep=dev->ep+ep_index;

    if(unlikely(user_buffer==NULL || size==0 || (ep->address&USB_DIR_IN)==0 || ep->type!=USB_ENDPOINT_XFER_BULK))
    {
        return -EINVAL;
    }

    buffer=kmalloc(size,GFP_KERNEL);

    if(unlikely(buffer==NULL))
    {
        return -ENOMEM;
    }

    usb_lock(dev);

    if(likely(dev->usb_interface))
    {
        status=usb_bulk_msg(dev->usb_dev,ep->pipe,buffer,size,&transferred,G69DDC_USB_TIMEOUT);
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock(dev);

    if(unlikely(status<0))
    {
        kfree(buffer);
        return status;
    }

    if(unlikely(copy_to_user(user_buffer,buffer,transferred)))
    {
        kfree(buffer);
        return -EFAULT;
    }

    kfree(buffer);

    return transferred;
}

static inline int g69ddc_usb_ioctl_bulk_msg_rcv(struct g69ddc_usb *dev,void __user *user_ptr)
{
 struct g69ddc_ioctl_usb_bulk_msg msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g69ddc_usb_bulk_msg_rcv(dev,msg.ep_index,(void __user*)msg.data,msg.size);
}

static inline int g69ddc_usb_ioctl_bulk_msg_snd(struct g69ddc_usb *dev,void __user *user_ptr)
{
 struct g69ddc_ioctl_usb_bulk_msg msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g69ddc_usb_bulk_msg_snd(dev,msg.ep_index,(void __user*)msg.data,msg.size);
}

static int g69ddc_usb_ioctl_set_led_mode(struct g69ddc_usb *dev,uint32_t led_mode)
{
 int status;

    switch(led_mode)
    {
        case USB_LED_MODE_OFF:
        case USB_LED_MODE_ON:
        case USB_LED_MODE_DIAG:
            dev_usb_lock(dev);

            if(led_mode!=dev->led_mode)
            {
                dev->led_mode=led_mode;
                status=g69ddc_usb_update_led_pattern(dev);
            }
            else
            {
                status=0;
            }

            dev_usb_unlock(dev);

            return status;

        default:
            return -EINVAL;
    }

    return g69ddc_usb_update_led_pattern(dev);
}

static int g69ddc_usb_ioctl_get_ep_info(struct g69ddc_usb *dev,struct g69ddc_ioctl_usb_ep __user *user_ep)
{
 uint32_t i;
 struct g69ddc_ioctl_usb_ep ep;

    for(i=0;i<dev->ep_count;i++)
    {
        ep.address=dev->ep[i].address;
        ep.interval=dev->ep[i].interval;
        ep.max_packet_size=dev->ep[i].max_packet_size;
        ep.type=dev->ep[i].type;

        if(unlikely(copy_to_user(user_ep,&ep,sizeof(ep))))
        {
            return -EFAULT;
        }

        user_ep++;
    }

    return dev->ep_count;
}

static int g69ddc_usb_ep_get_complete_buffer(struct g69ddc_usb *dev,uint32_t ep_index,uintptr_t *user_buffer)
{
 struct g69ddc_usb_ep_buffer *buffer;
 struct g69ddc_usb_ep *ep;
 unsigned long flags;
 int status;
 DEFINE_WAIT(wait);

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    ep=&dev->ep[ep_index];

    for(;;)
    {
        usb_lock(dev);

        if(unlikely(dev->usb_interface==NULL))
        {
            usb_unlock(dev);
            return -ENODEV;
        }

        if(unlikely(dev->base_dev.state.s.power==0))
        {
            usb_unlock(dev);
            return -EPERM;
        }

        spin_lock_irqsave(&ep->spinlock,flags);

        if(likely(!list_empty(&ep->buffer_list)))
        {
            buffer=list_first_entry(&ep->buffer_list,typeof(*buffer),list_entry);

            if(buffer->idle)
            {
                if(likely(buffer->status==0))
                {
                    *user_buffer=buffer->user_buffer;
                    status=buffer->bytes_transferred;
                }
                else
                {
                    status=buffer->status;
                }

                spin_unlock_irqrestore(&ep->spinlock,flags);
                usb_unlock(dev);
                return status;
            }
        }

        prepare_to_wait(&ep->wait_queue,&wait,TASK_INTERRUPTIBLE);

        spin_unlock_irqrestore(&ep->spinlock,flags);

        usb_unlock(dev);

        schedule();

        finish_wait(&ep->wait_queue,&wait);

        if(unlikely(signal_pending(current)))
        {
            return -ERESTARTSYS;
        }
    }
}

static inline int g69ddc_usb_ioctl_ep_get_complete_buffer(struct g69ddc_usb *dev,void __user *user_ptr)
{
 union g69ddc_ioctl_usb_ep_complete_buffer tmp;
 uintptr_t user_buffer;
 int status;

    if(unlikely(copy_from_user(&tmp.ep_index,user_ptr,sizeof(tmp.ep_index))))
    {
        return -EFAULT;
    }

    status=g69ddc_usb_ep_get_complete_buffer(dev,tmp.ep_index,&user_buffer);

    if(likely(status>=0))
    {
        if(unlikely(copy_to_user((void __user*)user_ptr,&user_buffer,sizeof(tmp.buffer))))
        {
            return -EFAULT;
        }
    }

    return status;
}

static long g69ddc_usb_ioctl(struct g69ddc_usb *dev,struct file *file,unsigned int cmd,unsigned long arg)
{
    switch(cmd)
    {
        case G69DDC_IOCTL_USB_EP_GET_COMPLETE_BUFFER:
            return g69ddc_usb_ioctl_ep_get_complete_buffer(dev,(void __user*)arg);

        case G69DDC_IOCTL_USB_EP_REQUEUE_BUFFER:
            return g69ddc_usb_ioctl_ep_bulk_requeue_buffer(dev,(void __user*)arg);

        case G69DDC_IOCTL_USB_CTRL_VENDOR_MSG_RCV:
            return g69ddc_usb_ioctl_ctrl_vendor_msg_rcv(dev,(void __user*)arg);

        case G69DDC_IOCTL_USB_CTRL_VENDOR_MSG_SND:
            return g69ddc_usb_ioctl_ctrl_vendor_msg_snd(dev,(void __user*)arg);

        case G69DDC_IOCTL_USB_BULK_MSG_RCV:
            return g69ddc_usb_ioctl_bulk_msg_rcv(dev,(void __user*)arg);

        case G69DDC_IOCTL_USB_BULK_MSG_SND:
            return g69ddc_usb_ioctl_bulk_msg_snd(dev,(void __user*)arg);

        case G69DDC_IOCTL_USB_EP_ADD_BUFFER:
            return g69ddc_usb_ioctl_ep_add_buffer(dev,(void __user*)arg);

        case G69DDC_IOCTL_USB_EP_RELEASE_BUFFERS:
            return g69ddc_usb_ioctl_ep_cancel_and_release_buffers(dev,(uint32_t)arg);

        case G69DDC_IOCTL_USB_EP_CANCEL_BUFFERS:
            return g69ddc_usb_ioctl_ep_cancel_buffers(dev,(uint32_t)arg);

        case G69DDC_IOCTL_USB_EP_CLEAR_HALT:
            return g69ddc_usb_ioctl_clear_halt(dev,(uint32_t)arg);

        case G69DDC_IOCTL_USB_EP_RESET:
            return g69ddc_usb_ioctl_reset_ep(dev,(uint32_t)arg);

        case G69DDC_IOCTL_USB_SET_LED_MODE:
            return g69ddc_usb_ioctl_set_led_mode(dev,(uint32_t)arg);

        case G69DDC_IOCTL_USB_GET_EP_INFO:
            return g69ddc_usb_ioctl_get_ep_info(dev,(void __user*)arg);

        case G69DDC_IOCTL_USB_GET_EP_COUNT:
            return dev->ep_count;

        default:
            return g69ddc_ioctl(&dev->base_dev,file,cmd,arg);
    }
}

#ifdef G69DDC_COMPAT

static inline int g69ddc_usb_ioctl_ep_get_complete_buffer_compat(struct g69ddc_usb *dev,void __user *user_ptr)
{
 union g69ddc_ioctl_usb_ep_complete_buffer_compat tmp;
 uintptr_t user_buffer;
 int status;

    if(unlikely(copy_from_user(&tmp.ep_index,user_ptr,sizeof(tmp.ep_index))))
    {
        return -EFAULT;
    }

    status=g69ddc_usb_ep_get_complete_buffer(dev,tmp.ep_index,&user_buffer);

    if(likely(status>=0))
    {
        if(unlikely(copy_to_user(user_ptr,&user_buffer,sizeof(tmp.buffer))))
        {
            return -EFAULT;
        }
    }

    return status;
}

static inline int g69ddc_usb_ioctl_ctrl_vendor_msg_rcv_compat(struct g69ddc_usb *dev,void __user *user_ptr)
{
 struct g69ddc_ioctl_usb_ctrl_vendor_msg_compat msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g69ddc_usb_ctrl_vendor_msg_rcv(dev,msg.request,msg.value,msg.index,compat_ptr(msg.data),msg.size);
}

static inline int g69ddc_usb_ioctl_ctrl_vendor_msg_snd_compat(struct g69ddc_usb *dev,void __user *user_ptr)
{
 struct g69ddc_ioctl_usb_ctrl_vendor_msg_compat msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g69ddc_usb_ctrl_vendor_msg_snd(dev,msg.request,msg.value,msg.index,compat_ptr(msg.data),msg.size);
}

static inline int g69ddc_usb_ioctl_bulk_msg_rcv_compat(struct g69ddc_usb *dev,void __user *user_ptr)
{
 struct g69ddc_ioctl_usb_bulk_msg_compat msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g69ddc_usb_bulk_msg_rcv(dev,msg.ep_index,compat_ptr(msg.data),msg.size);
}

static inline int g69ddc_usb_ioctl_bulk_msg_snd_compat(struct g69ddc_usb *dev,void __user *user_ptr)
{
 struct g69ddc_ioctl_usb_bulk_msg_compat msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g69ddc_usb_bulk_msg_snd(dev,msg.ep_index,compat_ptr(msg.data),msg.size);
}

static inline int g69ddc_usb_ioctl_ep_add_buffer_compat(struct g69ddc_usb *dev,void __user *user_ptr)
{
 struct g69ddc_ioctl_usb_add_ep_buffer_compat tmp;

    if(unlikely(copy_from_user(&tmp,user_ptr,sizeof(tmp))))
    {
        return -EFAULT;
    }

    return g69ddc_usb_ep_add_buffer(dev,tmp.ep_index,(uintptr_t)compat_ptr(tmp.buffer),tmp.size);
}

static long g69ddc_usb_ioctl_compat(struct g69ddc_usb *dev,struct file *file, unsigned int cmd, unsigned long arg)
{
    switch(cmd)
    {
        case G69DDC_IOCTL_USB_EP_GET_COMPLETE_BUFFER_COMPAT:
            return g69ddc_usb_ioctl_ep_get_complete_buffer_compat(dev,compat_ptr(arg));

        case G69DDC_IOCTL_USB_EP_REQUEUE_BUFFER:
            return g69ddc_usb_ioctl_ep_bulk_requeue_buffer(dev,compat_ptr(arg));

        case G69DDC_IOCTL_USB_CTRL_VENDOR_MSG_RCV_COMPAT:
            return g69ddc_usb_ioctl_ctrl_vendor_msg_rcv_compat(dev,compat_ptr(arg));

        case G69DDC_IOCTL_USB_CTRL_VENDOR_MSG_SND_COMPAT:
            return g69ddc_usb_ioctl_ctrl_vendor_msg_snd_compat(dev,compat_ptr(arg));

        case G69DDC_IOCTL_USB_BULK_MSG_RCV_COMPAT:
            return g69ddc_usb_ioctl_bulk_msg_rcv_compat(dev,compat_ptr(arg));

        case G69DDC_IOCTL_USB_BULK_MSG_SND_COMPAT:
            return g69ddc_usb_ioctl_bulk_msg_snd_compat(dev,compat_ptr(arg));

        case G69DDC_IOCTL_USB_EP_ADD_BUFFER_COMPAT:
            return g69ddc_usb_ioctl_ep_add_buffer_compat(dev,compat_ptr(arg));

        case G69DDC_IOCTL_USB_EP_RELEASE_BUFFERS:
            return g69ddc_usb_ioctl_ep_cancel_and_release_buffers(dev,(uint32_t)arg);

        case G69DDC_IOCTL_USB_EP_CANCEL_BUFFERS:
            return g69ddc_usb_ioctl_ep_cancel_buffers(dev,(uint32_t)arg);

        case G69DDC_IOCTL_USB_EP_CLEAR_HALT:
            return g69ddc_usb_ioctl_clear_halt(dev,(uint32_t)arg);

        case G69DDC_IOCTL_USB_EP_RESET:
            return g69ddc_usb_ioctl_reset_ep(dev,(uint32_t)arg);

        case G69DDC_IOCTL_USB_SET_LED_MODE:
            return g69ddc_usb_ioctl_set_led_mode(dev,(uint32_t)arg);

        case G69DDC_IOCTL_USB_GET_EP_INFO:
            return g69ddc_usb_ioctl_get_ep_info(dev,compat_ptr(arg));

        case G69DDC_IOCTL_USB_GET_EP_COUNT:
            return dev->ep_count;

        default:
            return g69ddc_ioctl_compat(&dev->base_dev,file,cmd,arg);
    }
}

#endif //G69DDC_COMPAT

static int g69ddc_usb_set_power_locked(struct g69ddc_usb *dev,uint8_t power)
{
 uint32_t i;
 int status;

    if(dev->base_dev.state.s.power==power)
    {
        return 0;
    }

    usb_lock_exclusive(dev);

    if(unlikely(dev->usb_interface==NULL))
    {
        usb_unlock_exclusive(dev);
        return -ENODEV;
    }

    if(power==0)
    {
        SET_STATE(&dev->base_dev,power=0);

        for(i=0;i<dev->ep_count;i++)
        {
            usb_kill_anchored_urbs(&dev->ep[i].anchor);
            wake_up(&dev->ep[i].wait_queue);
        }
    }

    status=usb_control_msg(dev->usb_dev,dev->snd_ctrl_pipe,
                       0xC7,USB_TYPE_VENDOR|USB_DIR_OUT,power,0,NULL,0,G69DDC_USB_TIMEOUT);

    usb_unlock_exclusive(dev);

    if(likely(status==0))
    {
        if(power)
        {
            SET_STATE(&dev->base_dev,power=power);
        }

        g69ddc_usb_update_led_pattern(dev);
    }

    return status;
}


static void g69ddc_usb_delete(struct g69ddc_usb *dev)
{
    usb_put_intf(dev->usb_interface);
    usb_put_dev(dev->usb_dev);
}

static int g69ddc_usb_open_locked(struct g69ddc_usb *dev)
{
    g69ddc_usb_update_led_pattern(dev);

    return 0;
}

static void g69ddc_usb_close_locked(struct g69ddc_usb *dev)
{
 uint32_t i;

    g69ddc_usb_set_power_locked(dev,0);

    usb_lock(dev);

    for(i=0;i<dev->ep_count;i++)
    {
        g69ddc_usb_ep_release_buffers(&dev->ep[i]);
    }

    usb_unlock(dev);
}

static int g69ddc_usb_check_temperature_locked(struct g69ddc_usb *dev,bool *critical_temperature)
{
 int16_t *temps;
 int status;

    temps=kmalloc(4*sizeof(int16_t),GFP_KERNEL);

    if(unlikely(temps==NULL))
    {
        dev_print_error(&dev->usb_dev->dev,"out of memory!");
        return -ENOMEM;
    }

    usb_lock(dev);

    if(likely(dev->usb_interface!=NULL))
    {
        status=usb_control_msg(dev->usb_dev,dev->rcv_ctrl_pipe,0xCB,USB_TYPE_VENDOR|USB_DIR_IN,0,0,temps,4*sizeof(int16_t),G69DDC_USB_TIMEOUT);

        if(likely(status>=4*sizeof(int16_t)))
        {
            *critical_temperature=(int16_t)le16_to_cpu(temps[0])>=USB_CRITICAL_TEMPERATURE_0 ||
                                  (int16_t)le16_to_cpu(temps[1])>=USB_CRITICAL_TEMPERATURE_1 ||
                                  (int16_t)le16_to_cpu(temps[2])>=USB_CRITICAL_TEMPERATURE_2;

            status=0;
        }
        else
        {
            status=-EIO;
        }
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock(dev);

    kfree(temps);

    return status;
}

static void g69ddc_usb_uninitialize(struct g69ddc_usb *dev)
{
}


static int g69ddc_usb_initialize(struct g69ddc_usb *dev)
{
 uint32_t i;

    dev->base_dev.fnc_delete=(g69ddc_fnc_delete)g69ddc_usb_delete;
    dev->base_dev.fnc_uninitialize=(g69ddc_fnc_uninitialize)g69ddc_usb_uninitialize;
    dev->base_dev.fnc_ioctl=(g69ddc_fnc_ioctl)g69ddc_usb_ioctl;
    dev->base_dev.fnc_open_locked=(g69ddc_fnc_open)g69ddc_usb_open_locked;
    dev->base_dev.fnc_close_locked=(g69ddc_fnc_close)g69ddc_usb_close_locked;
    dev->base_dev.fnc_set_power_locked=(g69ddc_fnc_set_power)g69ddc_usb_set_power_locked;
    dev->base_dev.fnc_check_temperature_locked=(g69ddc_fnc_check_temperature)g69ddc_usb_check_temperature_locked;

#ifdef G69DDC_COMPAT
    dev->base_dev.fnc_ioctl_compat=(g69ddc_fnc_ioctl)g69ddc_usb_ioctl_compat;
#endif

    dev->led_mode=USB_LED_MODE_DIAG;

    dev->base_dev.vid_pid[0]=USB_VENDOR_ID;
    dev->base_dev.vid_pid[1]=USB_PRODUCT_ID;

    init_rwsem(&dev->rw_sem);

    for(i=0;i<dev->ep_count;i++)
    {
        init_usb_anchor(&dev->ep[i].anchor);
        INIT_LIST_HEAD(&dev->ep[i].buffer_list);
        spin_lock_init(&dev->ep[i].spinlock);
        init_waitqueue_head(&dev->ep[i].wait_queue);
    }

    return 0;
}

static int g69ddc_usb_setup(struct g69ddc_usb *dev)
{
    g69ddc_usb_update_led_pattern(dev);

    return 0;
}

static int g69ddc_usb_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
 int status;
 struct g69ddc_usb *dev;
 struct usb_device *usb_dev;
 uint8_t i,ep_count;
 uint8_t ep_address;
 struct usb_endpoint_descriptor *ep_desc;
 int j;

    usb_dev=interface_to_usbdev(interface);

    dev_print_info(&usb_dev->dev,"probe");

    if(unlikely(id->idVendor!=USB_VENDOR_ID || id->idProduct!=USB_PRODUCT_ID))
    {
        dev_print_error(&usb_dev->dev,"unsupported device!");
        return -ENODEV;
    }

    if(unlikely(usb_dev->speed<USB_SPEED_HIGH))
    {
        dev_print_error(&usb_dev->dev,"device is not running at supported speed!");
        return -ENODEV;
    }

    ep_count=interface->cur_altsetting->desc.bNumEndpoints;

    dev=kzalloc(sizeof(*dev)+sizeof(*dev->ep)*ep_count,GFP_KERNEL);

    if(unlikely(dev==NULL))
    {
        print_error("out of memory!");
        return -ENOMEM;
    }

    dev->ep=(struct g69ddc_usb_ep*)(dev+1);

    status=usb_control_msg(usb_dev,usb_rcvctrlpipe(usb_dev,0),0xCC,USB_TYPE_VENDOR|USB_DIR_IN,0,0,dev->base_dev.fw,3*sizeof(uint16_t),G69DDC_USB_TIMEOUT);

    if(unlikely(status<0))
    {
        kfree(dev);
        dev_print_error(&usb_dev->dev,"failed to read device information!");
        return -ENODEV;
    }

    status/=sizeof(uint16_t);

    for(j=0;j<status;j++)
    {
        dev->base_dev.fw[j]=le16_to_cpu(dev->base_dev.fw[j]);
    }

    if(unlikely(usb_dev->speed<USB_SPEED_SUPER))
    {
        dev_print_warn(&usb_dev->dev,"device is not running at its maximum speed!");
        dev->base_dev.bus_type=BUS_TYPE_USB2;
    }
    else
    {
        dev->base_dev.bus_type=BUS_TYPE_USB3;
    }

    dev->usb_dev=usb_get_dev(usb_dev);
    dev->base_dev.p_dev=&usb_dev->dev;
    dev->usb_interface=usb_get_intf(interface);
    dev->ep_count=ep_count;
    dev->base_dev.p_dev=&usb_dev->dev;

    usb_set_intfdata(interface,dev);

    dev->snd_ctrl_pipe=usb_sndctrlpipe(usb_dev,0);
    dev->rcv_ctrl_pipe=usb_rcvctrlpipe(usb_dev,0);

    for(i=0;i<ep_count;i++)
    {
        ep_desc=&interface->cur_altsetting->endpoint[i].desc;

        ep_address=ep_desc->bEndpointAddress;

        dev->ep[i].dev=dev;
        dev->ep[i].address=ep_address;
        dev->ep[i].interval=ep_desc->bInterval;
        dev->ep[i].type=usb_endpoint_type(ep_desc);
        dev->ep[i].max_packet_size=usb_endpoint_maxp(ep_desc);

        switch(dev->ep[i].type)
        {
            case USB_ENDPOINT_XFER_CONTROL:
                dev->ep[i].pipe=(ep_address&USB_DIR_IN)?usb_rcvctrlpipe(usb_dev,ep_address):usb_sndctrlpipe(usb_dev,ep_address);
                break;

            case USB_ENDPOINT_XFER_ISOC:
                dev->ep[i].pipe=(ep_address&USB_DIR_IN)?usb_rcvisocpipe(usb_dev,ep_address):usb_sndisocpipe(usb_dev,ep_address);
                break;

            case USB_ENDPOINT_XFER_BULK:
                dev->ep[i].pipe=(ep_address&USB_DIR_IN)?usb_rcvbulkpipe(usb_dev,ep_address):usb_sndbulkpipe(usb_dev,ep_address);
                break;

            case USB_ENDPOINT_XFER_INT:
                dev->ep[i].pipe=(ep_address&USB_DIR_IN)?usb_rcvintpipe(usb_dev,ep_address):usb_sndintpipe(usb_dev,ep_address);
                break;
        }
    }

    status=g69ddc_initialize(&dev->base_dev);

    if(unlikely(status))
    {
        usb_set_intfdata(interface,NULL);
        usb_put_intf(dev->usb_interface);
        usb_put_dev(dev->usb_dev);
        kfree(dev);
        return status;
    }

    status=g69ddc_usb_initialize(dev);

    if(unlikely(status))
    {
        g69ddc_uninitialize(&dev->base_dev);
        usb_set_intfdata(interface,NULL);
        usb_put_intf(dev->usb_interface);
        usb_put_dev(dev->usb_dev);
        kfree(dev);
        return status;
    }

    status=g69ddc_register(&dev->base_dev);

    if(unlikely(status<0))
    {
        g69ddc_delete(&dev->base_dev);
        return status;
    }

    g69ddc_usb_setup(dev);

    return 0;
}

static void g69ddc_usb_disconnect(struct usb_interface *interface)
{
 struct g69ddc_usb *dev;

    dev=usb_get_intfdata(interface);

    dev_print_info(&dev->usb_dev->dev,"disconnect");

    usb_set_intfdata(interface,NULL);

    if(dev)
    {
        dev_usb_lock(dev);

        SET_STATE(&dev->base_dev,connected=0);

        g69ddc_usb_set_power_locked(dev,0);

        dev_usb_unlock(dev);

        g69ddc_cancel_wd_work(&dev->base_dev);

        g69ddc_unregister(&dev->base_dev);

        usb_lock_exclusive(dev);

        dev->usb_interface=NULL;

        usb_unlock_exclusive(dev);
    }
}


static int g69ddc_usb_suspend(struct usb_interface *interface,pm_message_t message)
{
 struct g69ddc_usb *dev;

    dev=usb_get_intfdata(interface);

    if(likely(dev!=NULL))
    {
        dev_usb_lock(dev);

        if(dev->base_dev.state.s.power)
        {
            dev->base_dev.state.s.suspended=1;
            g69ddc_usb_set_power_locked(dev,0);
        }

        dev_usb_unlock(dev);
    }

    return 0;
}

static int g69ddc_usb_resume(struct usb_interface *interface)
{
 struct g69ddc_usb *dev;

    dev=usb_get_intfdata(interface);

    if(likely(dev!=NULL))
    {
        g69ddc_usb_setup(dev);
    }

    return 0;
}



static struct usb_device_id g69ddc_usb_id_table[]=
{
    {USB_DEVICE(USB_VENDOR_ID,USB_PRODUCT_ID)},
    {}
};

MODULE_DEVICE_TABLE(usb,g69ddc_usb_id_table);

static struct usb_driver g69ddc_usb_driver=
{
    .name=G69DDC_MODULE_NAME,
    .probe=g69ddc_usb_probe,
    .disconnect=g69ddc_usb_disconnect,
    .id_table=g69ddc_usb_id_table,

    .suspend=g69ddc_usb_suspend,
    .resume=g69ddc_usb_resume,
    .reset_resume=g69ddc_usb_resume
};

#undef USB_DEV

#endif //G69DDC_USB



static int g69ddc_fo_open(struct inode *node, struct file *file)
{
 uint32_t index;
 int status;
 struct g69ddc *dev;

    index=iminor(file->f_path.dentry->d_inode);

    if(index>=G69DDC_MAX_DEVICE_COUNT)
    {
        return -ENODEV;
    }

    reg_lock();

    dev=registered_devices[index];

    if(dev==NULL)
    {
        reg_unlock();
        return -ENODEV;
    }

    get_device(&dev->dev);

    dev_lock(dev);

    if(dev->open)
    {
        dev_unlock(dev);
        reg_unlock();
        put_device(&dev->dev);
        return -EBUSY;
    }

    dev->open=true;

    status=dev->fnc_open_locked(dev);

    if(status!=0)
    {
        dev->open=false;
        dev_unlock(dev);
        reg_unlock();
        put_device(&dev->dev);
        return status;
    }
    file->private_data=dev;

    dev_unlock(dev);

    reg_unlock();

    return 0;
}

static int g69ddc_fo_release(struct inode *node, struct file *file)
{
 struct g69ddc *dev;

    dev=file->private_data;

    if(dev!=NULL)
    {
        g69ddc_cancel_wd_work(dev);

        dev_lock(dev);

        dev->open=false;

        dev->fnc_close_locked(dev);

        wake_up_all(&dev->state_wait_queue);

        dev_unlock(dev);

        put_device(&dev->dev);

        file->private_data=NULL;
    }

    return 0;
}


static long g69ddc_fo_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    if(unlikely(file->private_data==NULL))
    {
        return -ENODEV;
    }

    return ((struct g69ddc*)file->private_data)->fnc_ioctl(file->private_data,file,cmd,arg);
}


#ifdef G69DDC_COMPAT
static long g69ddc_fo_ioctl_compat(struct file *file, unsigned int cmd, unsigned long arg)
{
    if(unlikely(file->private_data==NULL))
    {
        return -ENODEV;
    }

    return ((struct g69ddc*)file->private_data)->fnc_ioctl_compat(file->private_data,file,cmd,arg);
}
#endif


static int __init g69ddc_init(void)
{
 int status;

    status=alloc_chrdev_region(&base_device_number,0,G69DDC_MAX_DEVICE_COUNT,G69DDC_MODULE_NAME);

    if(status!=0)
    {
        print_error("could not register a range of device numbers");
        return status;
    }

    memset(registered_devices,0,sizeof(registered_devices));
    mutex_init(&reg_mutex);

    status=class_register(&g69ddc_class);

    if(status<0)
    {
        print_error("failed to register class!");
        goto error_unregister_chrdev_region;
    }

#ifdef G69DDC_USB

    status=usb_register(&g69ddc_usb_driver);

    if(status!=0)
    {
        print_error("failed to register usb driver!");
        goto error_class_unregister;
    }

#endif //G69DDC_USB

    return 0;

    error_class_unregister:
        class_unregister(&g69ddc_class);

    error_unregister_chrdev_region:
        unregister_chrdev_region(base_device_number,G69DDC_MAX_DEVICE_COUNT);

    return status;
}

static void __exit g69ddc_exit(void)
{
#ifdef G69DDC_USB
    usb_deregister(&g69ddc_usb_driver);
#endif //G69DDC_USB

    class_unregister(&g69ddc_class);
    unregister_chrdev_region(base_device_number,G69DDC_MAX_DEVICE_COUNT);

    mutex_destroy(&reg_mutex);
}

module_init(g69ddc_init);
module_exit(g69ddc_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR(G69DDC_AUTHOR);
MODULE_DESCRIPTION(G69DDC_DESCRIPTION);
